/*
 * Device Driver for DataTranslation DT2801
 *
 * Analog-Digital-Converter subsystem
 *
 */

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/timer.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifndef MODULE
#error "dt2801_adc must be compiled as a module"
#endif

#include "dt2801.h"
#include "adc.h"

static ADC_state adc_cf;

/* routines for faked driver operation, allows to test the driver
   when no hardware is present ( fake=1 as argument, to activate) */

static struct timer_list DT2801_timer_list;
static int fake_counter = 0;
static int stop_req = 0;
static int running = 0;

static void DT2801_start_timer(void);

static void DT2801_timer_interrupt(unsigned long unused)
{
        del_timer(&DT2801_timer_list);
        if (dt2801->fake != 0) {
                short *b = (short *) adc_cf.dmabuf;

                b[fake_counter] = fake_counter%5;
                fake_counter++;
		DT2801_start_timer();
        }
}

static void DT2801_start_timer(void)
{
        if (fake_counter >= adc_cf.conv) {
		running = 0;
                return;
        }
	if (stop_req != 0) {
		running = 0;
                return;
	}

        init_timer(&DT2801_timer_list);
        DT2801_timer_list.expires = jiffies + 20;
        DT2801_timer_list.function = DT2801_timer_interrupt;
        add_timer(&DT2801_timer_list);
}

/* end of fake routines */


static int allocate_adc_buffer(int bufsize)
{
        if ( bufsize == 0 || bufsize < 0 ) {
                printk("dt2801_adc: allocating DMA-buffer,"
                       " illegal bufsize %d\n", bufsize);
                return -(EINVAL);
        }
        if (bufsize > DT2801_MAX_DMA_SIZE) {
                printk("dt2801_adc: Requested DMA buffer size,"
                       " %d, too large\n", bufsize );
                return -(EINVAL);
        }

        /* reuse the buffer if it is large enough */
        if (bufsize <= adc_cf.dmabufsize ) {
                return 0;
        }
 
        if (adc_cf.dmabuf != NULL) {
                printk("dt2801_adc: freeing dmabuf\n");
                kfree(adc_cf.dmabuf);
                adc_cf.dmabufsize = 0;
                adc_cf.dmabuf = NULL;
        }

        /* Allocate buffer for DMA transfers */
        adc_cf.dmabuf = kmalloc( bufsize, GFP_KERNEL | GFP_DMA );
        if (adc_cf.dmabuf == NULL) {
                printk( "dt2801_adc: Unable to allocate DMA buffer\n" );
                return -(ENOMEM);
        }
        adc_cf.dmabufsize = bufsize;

        return 0;
}

static void clear_adc_buffer(void)
{
        short *b = (short *) adc_cf.dmabuf;
        int i;

        for (i=0; i<adc_cf.dmabufsize; i++) {
                b[i] = -1;
        }
}

static int get_currnt(void)
{
        short *b = (short *) adc_cf.dmabuf;
        int i;

	for (i=adc_cf.cnt; i<adc_cf.conv; i++) {
		if (b[i] == -1) {
                        break;
                }
        }

        adc_cf.cnt = i;

        return i;
}

static int dt2801_set_adc_params(Dt2801_ADCP_t *adcp)
{                
        int stat;

        dt2801_writecommand(DT_C_SET_AD);

        dt2801_writedata(adcp->gain);
        dt2801_writedata(adcp->startch);
        dt2801_writedata(adcp->endch);
        dt2801_writedata2(adcp->conv);

        stat = dt2801_checkforerror();
        if (stat!=0) {
                return stat;
        }

        adc_cf.last = 0;
        adc_cf.gain = adcp->gain;
        adc_cf.startch = adcp->startch;
        adc_cf.endch = adcp->endch;
        adc_cf.conv = adcp->conv;
        adc_cf.cnt = 0;
        
	if ((stat = allocate_adc_buffer(adc_cf.conv*sizeof(short)))!= 0) {
		return stat;
	}
	
        return 0;
}

static int dt2801_adc_ioctl_a(struct inode *inode, struct file *filp,
			      unsigned int cmd, unsigned long arg)
{
	ADCP* adcp = (ADCP*) filp->private_data;

        switch (cmd) {
        case DT2801_IOCTL_ADC_GET_POLARITY: {
                int stat;
                
                stat = verify_area(VERIFY_WRITE, (char *)arg, sizeof(int));
                if (stat != 0) {
                        return stat;
                }
                put_user(adc_cf.polarity, (int*)arg);
        }
        break;
	case DT2801_IOCTL_ADC_SET_GAIN: {
		int gain = (int) arg;
		adcp->gain = gain;
	}
	break;
	case DT2801_IOCTL_ADC_SETEXTTRIGGER: {
		int exttrig = (int) arg;
		adcp->exttrigger = exttrig;
	}
	break;
        default: {
                return dt2801_ioctl(inode, filp, cmd, arg);
        }
        }

        return 0;
}

static int dt2801_adc_ioctl(struct inode *inode, struct file *filp,
                            unsigned int cmd, unsigned long arg)
{
	ADCP* adcp = (ADCP*) filp->private_data;

        switch (cmd) {
        case DT2801_IOCTL_ADC_SET_PARAMS: {
                Dt2801_ADCP_t uadcp;

                memcpy_fromfs(&uadcp, (char*)arg, sizeof(uadcp));
                adcp->startch = uadcp.startch;
                adcp->endch = uadcp.endch;
                adcp->gain = uadcp.gain;
                adcp->conv = uadcp.conv;
                return dt2801_set_adc_params(&uadcp);
        }
        break;
        
        /* ioctl SET_ADC_PARAMS has to be called before every call to
           BLOCKSTART */
        case DT2801_IOCTL_ADC_BLOCKSTART: {
                if (dt2801->fake != 0) {
                        /* Start timer to fill the buffer */
                        if (running == 1) {
                                printk("dt2801: can't start, still running\n");
                                return -1;
                        }
                        clear_adc_buffer();
                        adc_cf.cnt = 0;
                
                        fake_counter = 0;
			running = 1;
			stop_req = 0;
                        DT2801_start_timer();
                } else {
                        clear_adc_buffer();
                        adc_cf.cnt = 0;

                        /* no 'printk's within cli()/sti() sections */
                        cli();
                        disable_dma(dt2801->dmachannel);
                        clear_dma_ff( dt2801->dmachannel );
                        set_dma_addr( dt2801->dmachannel, 
                                      (unsigned long)adc_cf.dmabuf );
                        set_dma_count( dt2801->dmachannel,
                                       adc_cf.dmabufsize );
                        set_dma_mode( dt2801->dmachannel, DMA_MODE_READ );
                        enable_dma(dt2801->dmachannel);
                        sti();
                        
                        if (adcp->exttrigger == 1) {
                                dt2801_writecommand(
                                                    DT_C_READ_AD | DT_MOD_DMA
                                                    | DT_MOD_EXTTRIG);
                        } else {
                                dt2801_writecommand(
                                                    DT_C_READ_AD | DT_MOD_DMA);
                        }
                        if (dt2801_checkforerror() != 0) {
                                printk("dt2801: blockstart-dma not ok\n");
                                return -1;
                        }
                        adc_cf.running = 1;
                }
                
        }
        break;
        case DT2801_IOCTL_ADC_BLOCKSTOP: {
		dt2801_do_stop();
		/* Stopping a command causes the error-flag to be set,
		   will have to clear it !! */
		dt2801_clearerror();
		if (dt2801->fake == 1) {
			stop_req = 1;
		} else {
			disable_dma(dt2801->dmachannel);
		}
	}
        break;
        case DT2801_IOCTL_ADC_RESIDUE: {
                int stat, residue;
                
                stat = verify_area(VERIFY_WRITE, (char *)arg, sizeof(int));
                
                if ( stat != 0) {
                        return stat;
                }
                
		cli();
                residue = get_dma_residue(dt2801->dmachannel);
		sti();
                
                put_user(residue, (int*)arg);
                
        }
        break;
	case DT2801_IOCTL_ADC_SETEXTCLOCK: {
		int extcl = (int) arg;
		adcp->extclock = extcl;
	}
	break;
	case DT2801_IOCTL_ADC_SETCONTINOUS: {
		int cont_flag;
		cont_flag = (int) arg;
	
		adcp->continous = cont_flag;
	}
	break;
        default: {
                return dt2801_adc_ioctl_a(inode, filp, cmd, arg);
        }
	}

    
        return 0;
}

static int adc_read_imm(struct inode *inode, struct file *filp,
                           char* buffer, int count)
{
	ADCP* adcp = (ADCP*) filp->private_data;
        int data;
        int stat;


        if (count != 4) {
		return -EINVAL;
        }

/*
dt2801_wait();
stat = inb_p(DT_STATUS);
printk("dt2801:  status-1 = 0x%0x\n", stat);
*/

        stat = dt2801_writecommand(DT_C_READ_ADIM);
        dt2801_writedata(adcp->gain);
        dt2801_writedata(adcp->startch);
        stat = dt2801_readdata2(&data);

        if (stat != 0) {
             printk("dt2801: stat = %d\n", stat);
             return -EIO;
        }

        count = sizeof(data);
        if ((stat = verify_area(VERIFY_WRITE, buffer, count)) != 0) {
                return stat;
        }
        memcpy_tofs(buffer, &data, count);

        return count;
}

static int adc_use_residue = 0;

static int adc_read_block(struct inode *inode, struct file *filp,
                          char* buffer, int count)
{
	/*
	ADCP* adcp = (ADCP*) filp->private_data;
	*/

        int residue;
        int d, stat;


        if (adc_cf.last == adc_cf.conv) {
                return -(EIO); /* eof ? */
        }
                
        if (dt2801->fake != 0) {
                adc_cf.currnt = get_currnt();
        } else {
		if (adc_use_residue == 1) {
			cli();
			residue = get_dma_residue(dt2801->dmachannel);
			/* only transfer complete 2-byte values  */
			adc_cf.currnt = adc_cf.conv - ((residue+1)/sizeof(short));
			sti();
		} else {
			adc_cf.currnt = get_currnt();
		}
        }
        
        /* conversion not started, generate error */
        if (adc_cf.currnt  == 0) {
                return -1;
        }
        
        if (adc_cf.currnt == adc_cf.conv) {
                adc_cf.currnt++;
        }
        
        d = adc_cf.currnt - adc_cf.last -1;
        if (d > count) {
                d = count;
        }
        if (d < 0) {
                return -(EINVAL);
        }
        
        if ((stat = verify_area(VERIFY_WRITE, buffer, d*sizeof(short))) != 0) {
                return stat;
        }
        
        memcpy_tofs(buffer, adc_cf.dmabuf+adc_cf.last*sizeof(short),
                    d*sizeof(short));
        
        adc_cf.last+=d;
        
        return d*sizeof(short);
}

int dt2801_adc_reset(void)
{
        adc_cf.gain  = -1;
        adc_cf.startch  = -1;
        adc_cf.endch  = -1;
        adc_cf.conv  = -1;

        return 0;

}

int dt2801_adc_cleanup(void)
{

        if (adc_cf.dmabuf != NULL) {
                kfree(adc_cf.dmabuf);
        }

        return 0;
}

/* May be given as parameters at load-time */
static char* adc_polarity = "unipolar";
static char* adc_inputmode = "differential";

int dt2801_adc_init()
{
        adc_cf.bits  = 12;

        if (!strcmp(adc_polarity, "unipolar")) {
		adc_cf.polarity  = DT2801_ADC_UNIPOLAR;
	} else if (!strcmp(adc_polarity, "bipolar")) {
		adc_cf.polarity  = DT2801_ADC_BIPOLAR;
	} else {
		printk("dt2801: invalid value for 'polarity': %s\n",
		       adc_polarity);
		return -(EINVAL);
	}

        if (!strcmp(adc_inputmode, "differential")) {
		adc_cf.inputmode  = DT2801_ADC_DIFFERENTIAL;
	} else if (!strcmp(adc_inputmode, "single-ended")) {
		adc_cf.inputmode  = DT2801_ADC_SINGLE_ENDED;
	} else {
		printk("dt2801: invalid value for 'inputmode': %s\n",
		       adc_inputmode);
		return -(EINVAL);
	}

        if (adc_cf.inputmode == DT2801_ADC_DIFFERENTIAL) {
		adc_cf.nchannels  = 8;
	} else {
		adc_cf.nchannels  = 16;
	}

        adc_cf.dmabufsize = 0;
        adc_cf.dmabuf     = NULL;

	adc_cf.inputmode  = DT2801_ADC_DIFFERENTIAL;
	printk("dt2801: ADC %s, %s\n", adc_polarity, adc_inputmode);

        return 0;
}

static void dt2801_adc_release(struct inode *inode, struct file *filp)
{
	ADCP* adcp = (ADCP*) filp->private_data;

        /* adc-usecount */
        kfree(adcp);

        dt2801_release(inode, filp);
}

static struct file_operations adc_fops_imm = {
        NULL,                /* seek */
        adc_read_imm,        /* read */
        NULL,                /* write */
        NULL,                /* readdir */
        NULL,                /* select */
        dt2801_adc_ioctl_a,  /* ioctl */
        NULL,                /* mmap */
        dt2801_adc_open,     /* open */
        dt2801_adc_release   /* release */
};

static struct file_operations adc_fops_block = {
        NULL,                /* seek */
        adc_read_block,      /* read */
        NULL,                /* write */
        NULL,                /* readdir */
        NULL,                /* select */
        dt2801_adc_ioctl,    /* ioctl */
        NULL,                /* mmap */
        dt2801_adc_open,     /* open */
        dt2801_adc_release   /* release */
};

int dt2801_adc_open(struct inode *inode, struct file *filp)
{
	int minor = MINOR(inode->i_rdev);
	int channel;
	ADCP* adcp;

	DT_PRINTKF("minor = %d\n", minor);

	if ((minor < 0 || minor > 15) && minor != DT2801_ADCX) {
		return -(ENODEV);
	}
        if ( (minor & 0x1f) == DT2801_ADCX) {
		channel = -1;
	} else {
		channel = minor & 0xf;
                if (channel < 0 || channel >= adc_cf.nchannels) {
			return -(ENODEV);
		}
	}

	DT_PRINTKF("channel = %d\n", channel);

        /* non-immediate */
        if (channel == -1 ) {
		filp->f_op = &adc_fops_block;
	} else {
		filp->f_op = &adc_fops_imm;
	}

	adcp = (ADCP*) kmalloc(sizeof(ADCP), GFP_KERNEL);
	if (adcp == NULL) {
		return -(ENOMEM);
	}

	filp->private_data = adcp;

        return 0;
}
