/*
 * Device Driver for DataTranslation DT2801
 *
 */

#ifndef _DT2801_H
#define _DT2801_H  1

#ifdef __KERNEL__

/* c.f. Linux Journal 3/96 p.13 */
#ifdef DEBUG_DT2801
#define DT_PRINTK(format, args...)  printk ("dt2801:" format, ## args)
#define DT_PRINTKF(format, args...)  printk ("dt2801:" "%s: " format, __FUNCTION__, ## args)
#else
#define DT_PRINTK(format, args...)  
#define DT_PRINTKF(format, args...) 
#endif

/* #define DT_PRINTK(format, args...) */

/* Hardware Configuration */
/* ====================== */

/*
 *  General
 */

/* Default major device number */
#ifndef DT2801_MAJOR
#define DT2801_MAJOR 62
#endif

/* I/O-port Baseaddress. Jumper selectable 10bit, last bit always
   zero (even address).

   Factory default.
 */
#ifndef DT2801_BASEADDR
#define DT2801_BASEADDR   0x2ec
#endif

/* Channel for DMA-transfers

   Factory default.
 */
#ifndef DT2801_DMACHANNEL
#define DT2801_DMACHANNEL    1
#endif

#define DT2801_MAX_DMA_SIZE (64 * 1024)

/* Ports */
#define DT2801_IOSIZE 2
#define DT_DATA_IN  (DT2801_BASEADDR)
#define DT_DATA_OUT (DT2801_BASEADDR)
#define DT_STATUS   (DT2801_BASEADDR+1)
#define DT_COMMAND  (DT2801_BASEADDR+1)


/* define's & typedef's */
/* ====================== */

/* Commands */
#define DT_C_RESET       0x0
#define DT_C_CLEAR_ERR   0x1
#define DT_C_READ_ERRREG 0x2
#define DT_C_SET_CLOCK   0x3

#define DT_C_TEST        0xb
#define DT_C_STOP        0xf

#define DT_C_SET_DIGIN   0x4
#define DT_C_SET_DIGOUT  0x5
#define DT_C_READ_DIG    0x6
#define DT_C_WRITE_DIG   0x7

#define DT_C_WRITE_DAIM  0x8
#define DT_C_SET_DA      0x9
#define DT_C_WRITE_DA    0xa

#define DT_C_READ_ADIM   0xc
#define DT_C_SET_AD      0xd
#define DT_C_READ_AD     0xe

/* Command modifiers (only used with read/write), EXTTRIG can be used 
   with some other commands.
*/
#define DT_MOD_DMA     (1<<4)
#define DT_MOD_CONT    (1<<5)
#define DT_MOD_EXTCLK  (1<<6)
#define DT_MOD_EXTTRIG (1<<7)

/* Bits in status register */
#define DT_S_DATA_OUT_READY   (1<<0)
#define DT_S_DATA_IN_FULL     (1<<1)
#define DT_S_READY            (1<<2)
#define DT_S_COMMAND          (1<<3)
#define DT_S_COMPOSITE_ERROR  (1<<7)

#define DT2801_TIMERCYCLES 2

/* 4 seems to be reasonable (see mapping of the minor device #'s) */
#define DT2801_MAX_BOARDS 4

/* Minor device #'s:

   bit  7,6 : board # 0-3
   bit  5   : subsystem:  0+O_RDONLY : A/D, 0+O_WRONLY: D/A,
                          1+{O_RDONLY | O_WRONLY}: D-I/O
   bit  4-0 : interpreted by the subsystems
*/

/* the 2 most significant bits represent the board # .
   (not used at the monent) */
#define DT2801_BOARD(minor) ( (minor & 0xc0) >> 6 )

/* bit 5 is 0 for a ADC or DAC, 1 for D-I/O; whether the device represents an
   ADC or an DAC depends on the mode (read-only or write-only). D-I/O lines
   can be opened for reading or writing, but not read-write.
*/
#define DT2801_IS_ADC(minor) ( (~minor & 0x20) )
#define DT2801_IS_DAC(minor) ( (~minor & 0x20) )
#define DT2801_IS_DIO(minor) ( ( minor & 0x20) )


/* Parameters for the whole board.
   Some are fixed through jumper settings. But I put these into
   the driver so they can be reported to programs. For now these
   will be given at module load-time. May make them changeable at
   runtime in the future. */


typedef struct {
	/* The clock (ticks) */
	int min;
	int max;
	int ticks;
} Clock_state;

typedef struct {

	int fake;

	int major;
	unsigned long baseaddr;
	int dmachannel;

	int inuse;
	int usecount;

	int inerror;
	int firstopen;
	int status;
	int dma_running;
	int exttrigger_expected;

        Clock_state clock;

} Dt2801_t;

extern Dt2801_t *dt2801;

#include "adc.h"
#include "dac.h"
#include "dio.h"


extern int dt2801_readdata(int*);
extern int dt2801_readdata2(int*);
extern int dt2801_writedata(unsigned int);
extern int dt2801_writedata2(unsigned int);

extern int dt2801_writecommand(int);

extern int dt2801_ioctl(struct inode*, struct file*,
                         unsigned int, unsigned long);
extern void dt2801_release(struct inode*, struct file*);
extern int dt2801_open(struct inode*, struct file*);
extern int dt2801_stop(void);
extern int dt2801_clearerror(void);
extern int dt2801_reset(void);
extern int dt2801_checkforerror(void);


#endif /* __KERNEL__ */

/* ioctl's */
#ifndef DT2801_MAGIC
#define DT2801_MAGIC    'D'
#endif

#define DT2801_IOCTL_STOP          _IO(DT2801_MAGIC, 0x01)
#define DT2801_IOCTL_SETCLOCK      _IOW(DT2801_MAGIC, 0x03, int)
#define DT2801_IOCTL_STATUS        _IOR(DT2801_MAGIC, 0x08, int)

#endif /* _DT2801_H */
