/*
 * Device Driver for DataTranslation DT2801
 *
 * kernel 2.0.x
 *
 */

static char rcs_revision[] = "$Revision: 0.5.1 $";


#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/timer.h>
#include <linux/ioport.h>

#include <asm/dma.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifdef MODULE
#include <linux/module.h>
#else
#error "dt2801mod must be compiled as a module"
#endif

#include "dt2801.h"
#include "adc.h"
#include "dac.h"
#include "dio.h"


Dt2801_t *dt2801 = NULL;

static int dt_timer_int = 0;
static struct timer_list dt_timer_list;

static struct wait_queue* dt2801_waitq = NULL;


static void dt2801_timer_interrupt(unsigned long unused)
{
	if (!dt_timer_int)  return;

	wake_up_interruptible(&dt2801_waitq);
}

static void dt2801_start_timer(void)
{
	init_timer(&dt_timer_list);
	dt_timer_list.expires = jiffies + DT2801_TIMERCYCLES;
	dt_timer_list.function = dt2801_timer_interrupt;
	dt_timer_int = 1;
	add_timer(&dt_timer_list);
}

void dt2801_wait(void)
{
	dt2801_start_timer();
 
	interruptible_sleep_on(&dt2801_waitq);
	dt_timer_int = 0;
	del_timer(&dt_timer_list);
}

/* Stop board, and do a dummy read to clear DataOutReady-bit */
int dt2801_do_stop(void)
{
	int stat, ignored;

	if (dt2801->fake == 1) {
		return 0;
	}

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: before stop: status = 0x%0x\n", stat); */
	outb_p(DT_C_STOP, DT_COMMAND);
	/* sleep a bit ? Wait for ready flag. */

		dt2801_wait();
	ignored = inb_p(DT_DATA_IN);

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: after stop: status = 0x%0x\n", stat); */
		dt2801_wait();
	stat = inb_p(DT_STATUS);
	/* printk("dt2801: after stop: status = 0x%0x\n", stat); */

	return 0;
}

int dt2801_do_reset(void)
{
	int stat, ignored;

	if (dt2801->fake == 1) {
		return 0;
	}

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: before reset: status = 0x%0x\n", stat); */
	/* stat != 0 if DataInFull-Flag is cleared */
	stat= ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
        if (!stat) {
		dt2801_wait();
		stat = inb_p(DT_STATUS);
		stat= ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
		if (!stat) {
			printk("dt2801: do_reset timed out. "
			       "DataInFull not cleared\n");
			printk("dt2801: status = 0x%0x\n", stat);
		        return -1;
		}
	}

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: rst-1 status = 0x%0x\n", stat); */
	/* stat != 0 if Ready-Flag is set */
	stat = stat & DT_S_READY;
        if (!stat) {
		dt2801_wait();
		stat = inb_p(DT_STATUS);
		stat = stat & DT_S_READY;
		if (!stat) {
			printk("dt2801: do_reset timed out. "
			       "Ready not set\n");
			printk("dt2801: status = 0x%0x\n", stat);
		        return -1;
		}
	}
	outb_p(DT_C_RESET, DT_COMMAND);

	/* This is basically readdata(), but completes if
	   Ready OR DataOutReady are set. */
	stat = inb_p(DT_STATUS);
	/* printk("dt2801: rst-2 status = 0x%0x\n", stat); */
	stat = stat & (DT_S_DATA_OUT_READY);
        if (!stat) {
		dt2801_wait();
		stat = inb_p(DT_STATUS);
	        /* printk("dt2801: rst-3 status = 0x%0x\n", stat); */
		stat = stat & ( DT_S_DATA_OUT_READY);
		if (!stat) {
			printk("dt2801: do_reset timed out. "
			       "Ready not set\n");
			printk("dt2801: status = 0x%0x\n", stat);
		        return -1;
		}
	}
		dt2801_wait();
	ignored = inb_p(DT_DATA_IN);
		dt2801_wait();
	stat = inb_p(DT_STATUS);
	/* printk("dt2801: after reset: status = 0x%0x\n", stat); */

	return 0;
}


/* These are the low-level routines:
   writecommand: write a command to the board
   writedata: write data byte
   readdata: read data byte

   The check the status. If the board is not ready the sleep by
   calling dt2801_wait() and try again. If the board is still not
   ready they return with a value != 0 */


/* Only checks DataOutReady-flag, not the Ready-flag as it is done
   in the examples of the manual. I don't see why this should be
   necessary. */
int dt2801_readdata(int* data)
{
	static int fake_data = 0;
	int stat = 0;

	if (dt2801->fake != 0) {
		fake_data = (fake_data + 100)%4096;
		*data = fake_data;
		return 0;
	}

	stat = inb_p(DT_STATUS);
	if (stat & DT_S_COMPOSITE_ERROR) {
            return -1;
	}
	stat = stat & DT_S_DATA_OUT_READY;

	if (!stat) {
		dt2801_wait();

		stat = inb_p(DT_STATUS);
		stat = stat & DT_S_DATA_OUT_READY;

		if (!stat) {
			printk("dt2801: readdata timed out. "
			       "DataOutReady not set\n");
			return -1;
		}

	}
	*data = inb_p(DT_DATA_IN);

	return 0;
}

int dt2801_readdata2(int *data)
{
	static int fake_data = 0;
	int lb, hb;

	if (dt2801->fake != 0) {
		fake_data = (fake_data + 1)%10;
		*data = fake_data;
		return 0;
	}

	if (dt2801_readdata(&lb) != 0) {
		return -1;
	}
	if (dt2801_readdata(&hb) != 0) {
		return -2;
	}

	*data =  (hb<<8)+lb;
	return 0;
}

int dt2801_writedata(unsigned int data)
{
	int stat = 0;

	if (dt2801->fake != 0) {
		return 0;
	}

	stat = inb_p(DT_STATUS);
	if (stat & DT_S_COMPOSITE_ERROR) {
            return -1;
	}
	stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;
	if (!stat) {
		dt2801_wait();
		stat = inb_p(DT_STATUS);
		stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;

		if (!stat) {
			printk("dt2801: writedata timed out. "
				"DataInFull not cleared\n");
			return -1;
		}
	}
	outb_p(data & 0xff, DT_DATA_IN);

	return 0;
}

int dt2801_writedata2(unsigned int data)
{

	if ( dt2801_writedata( data & 0xff) != 0) {
		return -1;
	}
	if ( dt2801_writedata( (data >> 8) & 0xff ) != 0) {
		return -1;
	}

	return 0;
}

int dt2801_writecommand(int command)
{
	int stat;

	if (dt2801->fake != 0) {
		return 0;
	}

	stat = inb_p(DT_STATUS);
	if (stat & DT_S_COMPOSITE_ERROR) {
        	printk("dt2801: writecommand : warning composite-error\n");
	}
	stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;

	/* stat != 0 if DataInFull-Flag is cleared */
	if (!stat) {
		dt2801_wait();

		stat = inb_p(DT_STATUS);
		stat = ((stat)^DT_S_DATA_IN_FULL) & DT_S_DATA_IN_FULL;

		if (!stat) {
			printk("dt2801: writecommand timed out. "
			       "DataInFull not cleared\n");
			return -1;
		}
	}

	stat = inb_p(DT_STATUS);
	stat = stat & DT_S_READY;

	if (!stat) {
		dt2801_wait();

		stat = inb_p(DT_STATUS);
		stat = stat & DT_S_READY;

		if (!stat) {
			printk("dt2801: writecommand timed out."
			       "Ready not set\n");
			return -1;
		}

	}
	outb_p(command, DT_COMMAND);
	return 0;
}

/* Check composite-error-flag. If set read error-register and clear
 */

/* Maybe reset board etc, if we fail to determine the error-code */
int dt2801_checkforerror(void)
{
	int stat;
	int err;

	if (dt2801->fake == 1) {
		return 0;
	}

	stat = inb_p(DT_STATUS);

        if (stat & DT_S_COMPOSITE_ERROR) {

		/* printk("dt2801: checkforerror: status = 0x%0x.\n", stat); */
		stat = dt2801_writecommand(DT_C_READ_ERRREG);
                if (stat != 0) {
			printk("dt2801: checkforerror: failed to read"
			       " error-reg\n");
			printk("dt2801: reload the driver.\n");
		}
		stat = dt2801_readdata2(&err);
                if (stat != 0) {
			printk("dt2801: checkforerror: failed to read"
			       " error-data\n");
			printk("dt2801: reload the driver.\n");
		}
		printk("dt2801: Error #%d.\n", err);

		/* Clear error-register and composite-error flag */
		stat = dt2801_writecommand(DT_C_CLEAR_ERR);
		if (stat != 0) {
			printk("dt2801: checkforerror: clear-error failed\n");
		}
			
		return err;
	}

        return 0;
}

int dt2801_wait_cmd_completed(void) {
	int stat;

	if (dt2801->fake == 1) {
		return 0;
	}

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: cmd-1: status = 0x%0x\n", stat); */
	if ( (stat & DT_S_READY) && !(stat & DT_S_COMPOSITE_ERROR)) {
		return 0;
	}
	dt2801_wait();
	stat = inb_p(DT_STATUS);
	/* printk("dt2801: cmd-2: status = 0x%0x\n", stat); */
	if ( (stat & DT_S_READY) && !(stat & DT_S_COMPOSITE_ERROR)) {
		return 0;
	}
	printk("dt2801: cmd not successfully completed\n");

	return -1;
}

int dt2801_clearerror(void) {
	int stat;

	if (dt2801->fake == 1) {
		return 0;
	}

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: clerrerr-1: status = 0x%0x\n", stat); */
	stat = dt2801_writecommand(DT_C_CLEAR_ERR);
        if (stat != 0) {
		dt2801_checkforerror();
		return -1;
	}
	stat = dt2801_wait_cmd_completed();
        if (stat != 0) {
		dt2801_checkforerror();
		return -1;
	}

	return 0;
}

int dt2801_reset(void)
{
	int stat;

	stat = dt2801_do_reset();
        if (stat != 0) {
               printk("dt2801: reset: do_reset failed\n");
        }

	/* defaults after RESET or power-on */
	dt2801_adc_reset();
	dt2801_dac_reset();
	dt2801_dio_reset();

	dt2801->clock.ticks = 200;  /* 200 ticks =  1ms for DT2801 */

	return 0;
}

static int dt2801_setclock(int period)
{
	int stat, err;

	if (dt2801->fake == 1) {
		return 0;
	}

	stat = inb_p(DT_STATUS);
        /* printk("dt2801: setclock status = 0x%0x\n", stat); */
	if (stat & DT_S_COMPOSITE_ERROR) {
		err =  dt2801_checkforerror();
                printk("dt2801: setclock err = 0x%0x\n", err);
                return -1;
        }
        /* printk("dt2801: setclock %d\n", period); */
	stat = dt2801_writecommand(DT_C_SET_CLOCK);
        if (stat != 0) {
		err =  dt2801_checkforerror();
		return -1;
	}
	dt2801_writedata2(period);
        if (stat != 0) {
		err =  dt2801_checkforerror();
		return -1;
	}

	stat = dt2801_wait_cmd_completed();
        if (stat != 0) {
		err =  dt2801_checkforerror();
		return -1;
	}

	return stat;
}

int dt2801_ioctl(struct inode *inode, struct file *filp,
		 unsigned int cmd, unsigned long arg)
{

	switch (cmd) {
	case DT2801_IOCTL_STOP: {
		dt2801_do_stop();
	}
	break;

	case DT2801_IOCTL_SETCLOCK: {
		int cl;
		cl = (int) arg;
		
		if ( dt2801_setclock(cl) != 0) {
			return -1;
		}
		dt2801->clock.ticks = cl;
	}
	break;
		
	default:
		printk("dt2801: unknown ioctl %u.\n", cmd);
		return -EINVAL;
	}
	
	return 0;
}

static int release_resources(void)
{
	dt2801_adc_cleanup();
	dt2801_dac_cleanup();
	dt2801_dio_cleanup();

	free_dma(dt2801->dmachannel);
	release_region(dt2801->baseaddr, DT2801_IOSIZE);

	return 0;
}


/* called by the dt281_xxx_release */
void dt2801_release(struct inode *inode, struct file *filp)
{

	dt2801->usecount--;

        MOD_DEC_USE_COUNT;

	DT_PRINTKF("%d\n", dt2801->usecount);
}

int dt2801_open(struct inode *inode, struct file *filp)
{
	int stat = 0;
	int minor = MINOR(inode->i_rdev);
	int mode = filp->f_flags & O_ACCMODE;
	int board = DT2801_BOARD(minor);

	MOD_INC_USE_COUNT;

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: open: status = 0x%0x\n", stat); */

        if (dt2801->firstopen == 1) {
         	if (dt2801_do_reset() != 0) {
	        	printk("dt2801_open: do_reset failed!?\n");
	        	MOD_DEC_USE_COUNT;
	        	return -1;
	        }
             dt2801->firstopen = 0;
        }

        if (board != 0) {
		printk("dt2801_open: no board #%d.\n", board);
		stat = -(ENODEV);
		goto fail1;
	}

	if (DT2801_IS_ADC(minor) || DT2801_IS_DAC(minor)) {
		if (DT2801_IS_ADC(minor) && mode == O_RDONLY) {
			stat = dt2801_adc_open(inode, filp);
		} else if (DT2801_IS_DAC(minor) && mode == O_WRONLY) {
			stat = dt2801_dac_open(inode, filp);

		} else {
			printk("dt2801_open: minor = %d, bad AD/DA mode.\n",
			       minor);
			stat = -(EPERM);
		}

	} else if (DT2801_IS_DIO(minor)) {
		stat = dt2801_dio_open(inode, filp);
   
	} else {
		printk("dt2801_open: invalid minor = 0x%02x.\n", minor);
		stat = -(ENODEV);
	}

fail1:
	if (stat < 0) {
		MOD_DEC_USE_COUNT;
	        return stat;
	}
	
        dt2801->usecount++;
	DT_PRINTKF("%d\n", dt2801->usecount);

	return stat;

}

struct file_operations dt2801_fops = {
	NULL,            /* seek */
	NULL,            /* read */
	NULL,            /* write */
	NULL,            /* readdir */
	NULL,            /* select */
	dt2801_ioctl,    /* ioctl */
	NULL,            /* mmap */
	dt2801_open,     /* open */
	dt2801_release   /* release */
};

#ifdef MODULE    /* only module version */


/*
 These can be set from the commandline, when loading the module.
*/
static int fake = 0;

static int major = DT2801_MAJOR;

static int io = DT2801_BASEADDR;
static int dma = DT2801_DMACHANNEL;

int init_module(void)
{
        int stat;

	printk("dt2801: DataTranslation-DT2801 driver, %s\n", rcs_revision);

	dt2801 = (Dt2801_t *) kmalloc( sizeof(Dt2801_t), GFP_KERNEL );
	if (dt2801 == NULL) {
		return -(ENOMEM);
	}

	dt2801->fake = fake;

	dt2801->major      = major;
	dt2801->baseaddr   = io;
	dt2801->dmachannel = dma;

	dt2801->inuse = 0;
	dt2801->usecount = 0;

	dt2801->firstopen = 1;
	dt2801->inerror = 0;
	dt2801->status = -1;
	dt2801->dma_running = 0;
	dt2801->exttrigger_expected = 0;


	if ( request_dma(dt2801->dmachannel, "dt2801") ) {
		printk("dt2801_init: Cannot request DMA %d.\n",
		       dt2801->dmachannel);
		goto fail1;
	}

	if (check_region(dt2801->baseaddr, DT2801_IOSIZE) != 0) {
		printk("dt2801: init_module: " 
		       "I/O ports 0x%lx to 0x%lx already in use.\n",
		       dt2801->baseaddr, dt2801->baseaddr + DT2801_IOSIZE-1);
		goto fail2;
	}

	request_region(dt2801->baseaddr, DT2801_IOSIZE, "dt2801");

	if (register_chrdev(dt2801->major, "dt2801", &dt2801_fops)) {
		printk("dt2801_init: Cannot register to major device %d.\n",
		       dt2801->major);
		goto fail3;
	} 


	/* Stop to make sure the board can accept the next command */
	dt2801_do_stop(); 

	if (dt2801_do_reset() != 0) {
		printk("dt2801_init: board init failed!?\n");
		goto fail3;
	}

	/* Initialize Clock, A/D, D/A, D-I/O */

	dt2801->clock.min = 2;
	dt2801->clock.max = 0xffff;

	if (dt2801_adc_init() != 0) {
		printk("dt2801_adc_init: failed\n");
		goto error;
	}
	if (dt2801_dac_init() != 0) {
		printk("dt2801_dac_init: failed\n");
		goto error;
	}
	if (dt2801_dio_init() != 0) {
		printk("dt2801_dio_init: failed\n");
		goto error;
	}

	/* Reset the board, via dt2801_reset(),
	   which will call the dt2801_xxx_reset() routines. */
	if (dt2801_reset() != 0) {
		printk("dt2801_init: reset not ok.\n");
		goto error;
	}
	
	printk("dt2801: init_module: " 
	       "fake(%d), major(%d), dma(%d), io(0x%lx).\n",
	       dt2801->fake, dt2801->major,
	       dt2801->dmachannel, dt2801->baseaddr );

	/* A module consisting of more than one object-file will
	   export all symbols (even when declared static). So we
	   explicitly remove the symbols from the table.
	   c.f. /usr/doc/modules/HOWTO-modularize. */
	register_symtab(0);

	stat = inb_p(DT_STATUS);
	/* printk("dt2801: finally: status = 0x%0x\n", stat); */

	return 0;

error:

	printk("dt2801: error during init, cleaning up.\n");
	dt2801_adc_cleanup();
	dt2801_dac_cleanup();
	dt2801_dio_cleanup();

fail3:
	unregister_chrdev(dt2801->major, "dt2801");

fail2:
	printk("dt2801: unloading driver.\n");
	release_region(dt2801->baseaddr, DT2801_IOSIZE);
fail1:
	free_dma(dt2801->dmachannel);

	kfree(dt2801);

	return -1;
}

void cleanup_module(void)
{
	/* As 'cleanup_mmodule()' should only be called when the
	   module is no longer in use, this is not strictly
	   necessary (backward compatibility with older kernels)*/

	if (MOD_IN_USE) {
		printk("dt2801: device busy, remove delayed\n");
		return;
	} else {

		release_resources();

		if (unregister_chrdev(dt2801->major, "dt2801") != 0) {
			printk("dt2801: cleanup_module failed\n");
		} else {
			printk("dt2801: cleanup succeeded\n");
		}
		kfree(dt2801);
	}
}

#endif /* MODULE */
